import time
import random
import traceback
import uuid

from multiprocessing.pool import ThreadPool
from multiprocessing import Lock


class TaskError(Exception):

    def __init__(self, error):
        super(TaskError, self).__init__()

        self.error = error


class TaskPool:

    def __init__(self, workers, callback):
        self._n_started = 0
        self.workers = workers if workers > 1 else 1
        self.running = [None for _ in range(0, workers)]
        self.mutex = Lock()
        self._guard_mutex = Lock()
        self.pool = ThreadPool(processes=self.workers)
        self.callback = callback
        self._payloads = {}

    def _is_full(self):
        return self._n_started == self.workers

    def _task_entered(self, task_id, args):
        self._guard_mutex.acquire()
        self._n_started += 1
        self._payloads[task_id] = args
        self._guard_mutex.release()

    def _task_exited(self, task_id, meta, result):
        self._guard_mutex.acquire()
        self._n_started -= 1
        del self._payloads[task_id]
        self._guard_mutex.release()

        if self.callback:
            self.callback(meta, result)

    def _run_task(self, task_id, fn, args, meta):
        self._task_entered(task_id, args)

        try:
            result = fn(*args)
        except Exception as e:

            traceback.print_exc()

            result = TaskError(error=e)

        self._task_exited(task_id, meta, result)

    def running_payloads(self):
        return [self._payloads[task_id] for task_id in self._payloads.keys()]

    def run_task(self, fn, args, meta):

        # don't overflow the task runner with tasks
        while self._is_full():
            time.sleep(random.randint(5, 15) / 10.0)

        # apply task
        self.pool.apply_async(TaskPool._run_task, (self, str(uuid.uuid4()), fn, args, meta))






import json
import random
import time
import traceback
import uuid

import boto3
from botocore.errorfactory import ClientError
from multiprocessing import Lock

from .utils.task_pool import TaskPool, TaskError


class EmptyTaskOutput(object):
    pass


class TaskRunner(object):

    def __init__(self, fn, in_queue_names, config):
        self.fn = fn
        self.in_queue_names = in_queue_names
        self.results = []
        self._result_mutex = Lock()
        self.out_queues = []
        self._out_queue_names = None

        self.domain = config['domain']
        self.name = config['name']
        self.aws_config = config['aws_config']
        self.workers = config.get('workers', 1)
        self.priority_levels = list(reversed(range(config.get('priorities', 1))))
        self.final = config.get('final', False)
        self.interval = config.get('interval', 0)
        self.ignore_none = config.get('ignore_none', True)

    def queue_name(self, priority):
        base_name = '%s-%s' % (self.domain, self.name)

        if not priority:
            return '%s.fifo' % base_name

        return '%s--p%d.fifo' % (base_name, priority)

    @property
    def out_queue_names(self):
        if self.final:
            return []

        if self._out_queue_names is not None:
            return self._out_queue_names

        self._out_queue_names = [self.queue_name(priority) for priority in self.priority_levels]

        return self._out_queue_names

    def set_workers(self, workers):
        self.workers = workers

    def in_queues(self):
        # no input queues?
        if not self.in_queue_names:
            return []

        sqs = self.sqs()

        retrieved = False
        retries = 0
        while not retrieved:
            in_queues = []
            try:
                for in_queue_name in self.in_queue_names:
                    in_queues.append(sqs.get_queue_by_name(QueueName=in_queue_name))
                    retrieved = True
            except ClientError as e:
                # only handle non-existing queue error
                if e.response.get('Error', {}).get('Code', '') != 'AWS.SimpleQueueService.NonExistentQueue':
                    raise

                # wait until input queues are created
                retrieved = False
                retries += 1
                if retries == 1:
                    print('Waiting for input queues...')

                time.sleep(2)

        if retries > 0:
            print('OK')

        return in_queues

    def sqs(self):
        return boto3.resource(
            'sqs',
            self.aws_config['region'],
            aws_access_key_id=self.aws_config['key'],
            aws_secret_access_key=self.aws_config['secret'],
        )

    def setup(self):
        sqs = self.sqs()

        self.out_queues = []
        for queue_name in self.out_queue_names:
            try:
                q = sqs.get_queue_by_name(QueueName=queue_name)
            except ClientError as e:
                # only handle non-existing queue error
                if e.response.get('Error', {}).get('Code', '') != 'AWS.SimpleQueueService.NonExistentQueue':
                    raise

                q = sqs.create_queue(QueueName=queue_name, Attributes={
                    'FifoQueue': 'true',
                    'VisibilityTimeout': '120',
                })

            self.out_queues = [q] + self.out_queues

    def _on_task_finish(self, task_meta, task_output):
        # add to results
        self._result_mutex.acquire()
        self.results.append(task_output)

        if not(
            ((task_output is None) and self.ignore_none) or
            (isinstance(task_output, EmptyTaskOutput))
        ):
            if (not self.final) and (type(task_output) != TaskError):
                # if priority is a callable,
                # it will be calculated from the extracted message
                if callable(task_meta.get('priority')):
                    task_meta['priority'] = task_meta['priority'](task_output)

                # also write to queues for next task to pick up
                try:
                    self.out_queues[task_meta['priority']].send_message(
                        MessageBody=json.dumps({
                            'meta': task_meta,
                            'value': task_output
                        }),
                        MessageDeduplicationId=str(uuid.uuid4()),
                        MessageGroupId='-'
                    )
                except:
                    traceback.print_exc()

        self._result_mutex.release()

    def _run(self, args, priority=0):
        # create the thread pool
        pool = TaskPool(self.workers, callback=self._on_task_finish)

        # get input queues
        in_queues = self.in_queues()

        while True:
            # receive messages
            messages = []
            for in_queue in in_queues:
                if messages:
                    break

                messages += in_queue.receive_messages(MaxNumberOfMessages=min(self.workers, 10))

            if not in_queues:
                pool.run_task(self.fn, args, meta={
                    'priority': priority
                })

                # interval could either be a function or a number
                if callable(self.interval):
                    _interval = self.interval()
                else:
                    _interval = self.interval

                time.sleep(_interval)

            elif messages:
                # get payloads from messages & delete them from sqs
                payloads = []
                for message in messages:
                    payloads.append(json.loads(message.body))
                    message.delete()

                # run tasks
                for payload in payloads:
                    pool.run_task(self.fn, (payload['value'], ) + args, payload['meta'])

            # yield any available results
            if self.results:
                self._result_mutex.acquire()

                _error = None
                for result in self.results:
                    if type(result) == TaskError:

                        _error = result.error
                    else:
                        yield result

                if _error:
                    raise _error

                self.results = []
                self._result_mutex.release()

    def run(self, args, priority=0, iterate=False):
        if iterate:
            return self._run(args, priority=priority)

        for _ in self._run(args, priority=priority):
            pass



import os
import sys
import random
import string
import time

from sqspipes.task_client import TaskClient

try:
    AWS_KEY = os.environ['AWS_KEY']
    AWS_SECRET = os.environ['AWS_SECRET']
    AWS_REGION = os.environ.get('AWS_REGION', 'us-west-2')
except KeyError:
    from sqspipes.tests_env import *


def _generate(max_size):
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(random.randint(1, max_size)))


def _reduce(value, keep='vowels'):
    vowels = ['a', 'e', 'i', 'o', 'u', ]
    if keep == 'vowels':
        result = [v for v in value if v in vowels]
    else:
        result = [v for v in value if v not in vowels]

    return value, ''.join(result)


def _count(data):
    value, vowels = data

    return value, len(vowels)


client = TaskClient(domain='word-counter', aws_key=AWS_KEY, aws_secret=AWS_SECRET, aws_region=AWS_REGION)

client.register_tasks([
    {'method': _generate, 'workers': 32, 'interval': 0.1},
    {'method': _reduce, 'workers': 2},
    {'method': _count, 'workers': 16}
])


def generate(workers):
    for res in client.run('_generate', args=(10, ), iterate=True, workers=workers):
        print(res)


def reduce(workers):
    for res in client.run('_reduce', iterate=True, workers=workers):
        print('%s -> %s' % res)


def count(workers):
    for result in client.run('_count', iterate=True, workers=workers):
        print('%s -> %d' % result)


try:
    n_workers = int(sys.argv[2])
except ValueError:
    n_workers = None

try:
    if sys.argv[1] == 'generate':
        generate(n_workers)
    elif sys.argv[1] == 'reduce':
        reduce(n_workers)
    elif sys.argv[1] == 'count':
        count(n_workers)
    else:
        raise ValueError('Invalid argument: must be one of generate, reduce or count')
except IndexError:
    raise ValueError('Script argument is required')



from distutils.core import setup

try:
    long_description = open('README.rst', encoding='utf8').read()
except IOError:
    long_description = ''
except TypeError:
    long_description = ''


setup(
    # Application name
    name="sqspipes",

    # Version number
    version="0.1.2",

    # Application author details
    author="Dimitris Papaspyros",
    author_email="dimitris@orfium.com",

    # Packages
    packages=["sqspipes", "sqspipes/utils"],

    # Include additional files into the package
    include_package_data=True,

    # Details
    url="https://github.com/dipapaspyros/sqspipes",
    license="LICENSE",
    description="A multi-worker pipe mechanism that uses AWS SQS",
    long_description=long_description,

    # Dependent packages (distributions)
    install_requires=[
        'boto3',
        'botocore',
    ],
)



from collections import Iterable
from types import FunctionType

import boto3

from .task import TaskRunner


class TaskClientError(ValueError):
    pass


class TaskClient(object):

    SQS_FIFO_AVAILABLE_AT = ('us-west-2', 'us-east-2', )

    def __init__(self, domain, aws_key, aws_secret, aws_region='us-west-2', ignore_none=True):
        """
        :param domain: A unique identifier of the application that uses the client (e.g my_awesome_app)
        :param aws_key: The AWS key that will be used. Should have full access to the SQS service.
        :param aws_secret: The AWS secret for the key used.
        :param aws_region: The AWS region. Defaults to us-west-2
        :param ignore_none: If the task output is None, ignore it. Defaults to True.
        """

        if aws_region not in self.SQS_FIFO_AVAILABLE_AT:
            raise TaskClientError(
                '`aws_region` must be one of %s - FIFO queues are not supported on other regions.' %
                ','.join(self.SQS_FIFO_AVAILABLE_AT)
            )

        self.aws_region = aws_region
        self.aws_key = aws_key
        self.aws_secret = aws_secret

        if not domain:
            raise TaskClientError('`domain` field is required.')

        self.domain = domain
        self.ignore_none = ignore_none

        # tasks are initially empty
        self.tasks = []

    def register_tasks(self, tasks):
        """
        :param tasks: Registers the tasks. Each task might be of the following form:
            {
                'method':
                    A callable object. This is the function that will actually be executed.
                    For all tasks except for the first one, the first argument of this method
                    will be the result of the previous task's method.

                'name':
                    The name of this tasks. Tasks can later be retrieved by their name.
                    If no name is provided, the method's name is automatically used.

                'workers':
                    The number of worker threads that will be processing messages in parallel.
                    Defaults to 1.

                'priorities':
                    The number of different priority levels, where 0 is the lowest possible priority.
                    Defaults to 1, maximum value is 16.

                'interval':
                    Number of seconds to wait between each execution.
                    This only applies to the first task.
                    Defaults to 0.
            }

            For clarity, if the defaults are used, instead of the dictionary described above,
            a method might only be passed instead for each one of the tasks.
        :return:
        """
        if not tasks:
            raise TaskClientError('`tasks` field is required.')

        if not isinstance(tasks, list):
            raise TaskClientError('`tasks` field must be a list of task configurations.')

        # initialize tasks
        self.tasks = []
        prev_task = None
        for idx, t in enumerate(tasks):
            task = self._parse_tasks(
                t,
                in_queue_names=prev_task.out_queue_names if prev_task else [],
                final=idx == len(tasks) - 1
            )
            self.tasks.append(task)
            prev_task = task

    def _parse_tasks(self, task, in_queue_names=None, final=False):
        if type(task) == FunctionType:
            task = {
                'method': task,
            }

        if not task.get('name'):
            task['name'] = task['method'].__name__

        return TaskRunner(
            fn=task['method'],
            in_queue_names=in_queue_names,
            config={
                'domain': self.domain,
                'name': task['name'],
                'workers': task.get('workers', 1),
                'aws_config': {
                    'region': self.aws_region,
                    'key': self.aws_key,
                    'secret': self.aws_secret,
                },
                'priorities': min(task.get('priorities', 1), 16),
                'interval': task.get('interval', 0),
                'final': final,
                'ignore_none': self.ignore_none,
            }
        )

    def get_task_by_name(self, task_name):
        """
        :param task_name: Retrieves a task by its name. The tasks index can also be used.
        :return:
        """
        if type(task_name) == int:
            return self.tasks[task_name]

        return [t for t in self.tasks if t.name == task_name][0]

    @property
    def sqs_client(self):
        """
        :return: An SQS client
        """
        return boto3.client(
            'sqs',
            self.aws_region,
            aws_access_key_id=self.aws_key,
            aws_secret_access_key=self.aws_secret,
        )

    @property
    def queues(self):
        """
        :return: A list of the queues created for this domain
        """
        return self.sqs_client.list_queues(
            QueueNamePrefix=self.domain
        ).get('QueueUrls', [])

    def delete(self):
        """
        :return: Deletes all of the queues for this particular domain
        """
        for queue in self.queues:
            self.sqs_client.delete_queue(
                QueueUrl=queue
            )

    def purge(self):
        """
        :return: Empties all of the queues for this particular domain
        """
        for queue in self.queues:
            self.sqs_client.purge_queue(
                QueueUrl=queue
            )

    def run(self, task_name, args=None, priority=0, iterate=False, workers=None):
        """
        Executes a specific task from its configuration
        :param task_name: The name of the task to execute (or it's index for unnamed tasks)
        :param args: Arguments for the executed tasks
        :param priority: Use one of the task's priority levels to prioritize before other executions
        :param iterate: If set to true, it returns an iterator that yields the results while the task is running
        :return: An iterator if iterate is set to True, None otherwise (should never stop unless interrupted)
        """

        if not self.tasks:
            raise TaskClientError('Tasks have not been initialized.')

        if args is None:
            args = ()

        # find task
        task = self.get_task_by_name(task_name)

        # overwrite #workers
        if workers is not None:
            task.set_workers(workers)

        # setup queues & other utils
        task.setup()

        # run the task
        return task.run(args, priority=priority, iterate=iterate)
